import os

from django.core.exceptions import SuspiciousFileOperation, ValidationError
from django.core.files.storage import default_storage
from django.utils.translation import ugettext_lazy as _
from jet_django.deps.rest_framework import serializers


class FileUploadSerializer(serializers.Serializer):
    file = serializers.FileField(write_only=True)
    path = serializers.CharField(write_only=True)
    filename = serializers.CharField(write_only=True, required=False)
    uploaded_path = serializers.CharField(read_only=True)
    uploaded_url = serializers.CharField(read_only=True)

    def validate(self, attrs):
        if attrs.get('filename') is None:
            attrs['filename'] = attrs['file'].name

        attrs['full_path'] = os.path.join(attrs['path'], attrs['filename'])

        try:
            default_storage.get_available_name(attrs['full_path'])
        except SuspiciousFileOperation:
            raise ValidationError(_('forbidden path'))

        return attrs

    def save(self, **kwargs):
        request = self.context.get('request', None)
        uploaded_path = default_storage.save(self.validated_data['full_path'], self.validated_data['file'])

        self.instance = {
            'uploaded_path': uploaded_path,
            'uploaded_url': request.build_absolute_uri(default_storage.url(uploaded_path))
        }



from jet_django.deps.rest_framework import serializers

from jet_django.models.menu_item import MenuSettings


class MenuSettingsSerializer(serializers.ModelSerializer):
    class Meta:
        model = MenuSettings
        fields = (
            'items',
            'date_add'
        )



from django.conf import settings

JET_BACKEND_API_BASE_URL = getattr(settings, 'JET_BACKEND_API_BASE_URL', 'https://api.jetadmin.io/api')
JET_BACKEND_WEB_BASE_URL = getattr(settings, 'JET_BACKEND_WEB_BASE_URL', 'https://app.jetadmin.io')
JET_DEMO = getattr(settings, 'JET_DEMO', False)
JET_REGISTER_TOKEN_ON_START = getattr(settings, 'JET_REGISTER_TOKEN_ON_START', True)

JET_REST_FRAMEWORK = getattr(settings, 'JET_REST_FRAMEWORK', {
    'DEFAULT_PAGINATION_CLASS': 'jet_django.deps.rest_framework.pagination.PageNumberPagination',
    'PAGE_SIZE': 25,
    'DEFAULT_AUTHENTICATION_CLASSES': (),
    'DEFAULT_FILTER_BACKENDS': (
        'jet_django.deps.django_filters.rest_framework.DjangoFilterBackend',
        'jet_django.deps.rest_framework.filters.OrderingFilter',
    ),
    # 'DEFAULT_THROTTLE_CLASSES': (
    #     'jet_django.deps.rest_framework.throttling.AnonRateThrottle',
    #     'jet_django.deps.rest_framework.throttling.UserRateThrottle'
    # ),
    # 'DEFAULT_THROTTLE_RATES': {
    #     'anon': '120/minute',
    #     'user': '480/minute'
    # },
    'DEFAULT_RENDERER_CLASSES': (
        'jet_django.deps.rest_framework.renderers.JSONRenderer',
    ),
    'EXCEPTION_HANDLER': 'jet_django.views.exception_handler.jet_exception_handler',
})



"""
The most important decorator in this module is `@api_view`, which is used
for writing function-based views with REST framework.

There are also various decorators for setting the API policies on function
based views, as well as the `@detail_route` and `@list_route` decorators, which are
used to annotate methods on viewsets that should be included by routers.
"""
from __future__ import unicode_literals

import types
import warnings

from django.utils import six

from jet_django.deps.rest_framework.views import APIView


def api_view(http_method_names=None, exclude_from_schema=False):
    """
    Decorator that converts a function-based view into an APIView subclass.
    Takes a list of allowed methods for the view as an argument.
    """
    http_method_names = ['GET'] if (http_method_names is None) else http_method_names

    def decorator(func):

        WrappedAPIView = type(
            six.PY3 and 'WrappedAPIView' or b'WrappedAPIView',
            (APIView,),
            {'__doc__': func.__doc__}
        )

        # Note, the above allows us to set the docstring.
        # It is the equivalent of:
        #
        #     class WrappedAPIView(APIView):
        #         pass
        #     WrappedAPIView.__doc__ = func.doc    <--- Not possible to do this

        # api_view applied without (method_names)
        assert not(isinstance(http_method_names, types.FunctionType)), \
            '@api_view missing list of allowed HTTP methods'

        # api_view applied with eg. string instead of list of strings
        assert isinstance(http_method_names, (list, tuple)), \
            '@api_view expected a list of strings, received %s' % type(http_method_names).__name__

        allowed_methods = set(http_method_names) | {'options'}
        WrappedAPIView.http_method_names = [method.lower() for method in allowed_methods]

        def handler(self, *args, **kwargs):
            return func(*args, **kwargs)

        for method in http_method_names:
            setattr(WrappedAPIView, method.lower(), handler)

        WrappedAPIView.__name__ = func.__name__
        WrappedAPIView.__module__ = func.__module__

        WrappedAPIView.renderer_classes = getattr(func, 'renderer_classes',
                                                  APIView.renderer_classes)

        WrappedAPIView.parser_classes = getattr(func, 'parser_classes',
                                                APIView.parser_classes)

        WrappedAPIView.authentication_classes = getattr(func, 'authentication_classes',
                                                        APIView.authentication_classes)

        WrappedAPIView.throttle_classes = getattr(func, 'throttle_classes',
                                                  APIView.throttle_classes)

        WrappedAPIView.permission_classes = getattr(func, 'permission_classes',
                                                    APIView.permission_classes)

        WrappedAPIView.schema = getattr(func, 'schema',
                                        APIView.schema)

        if exclude_from_schema:
            warnings.warn(
                "The `exclude_from_schema` argument to `api_view` is deprecated. "
                "Use the `schema` decorator instead, passing `None`.",
                DeprecationWarning
            )
            WrappedAPIView.exclude_from_schema = exclude_from_schema

        return WrappedAPIView.as_view()
    return decorator


def renderer_classes(renderer_classes):
    def decorator(func):
        func.renderer_classes = renderer_classes
        return func
    return decorator


def parser_classes(parser_classes):
    def decorator(func):
        func.parser_classes = parser_classes
        return func
    return decorator


def authentication_classes(authentication_classes):
    def decorator(func):
        func.authentication_classes = authentication_classes
        return func
    return decorator


def throttle_classes(throttle_classes):
    def decorator(func):
        func.throttle_classes = throttle_classes
        return func
    return decorator


def permission_classes(permission_classes):
    def decorator(func):
        func.permission_classes = permission_classes
        return func
    return decorator


def schema(view_inspector):
    def decorator(func):
        func.schema = view_inspector
        return func
    return decorator


def action(methods=None, detail=None, url_path=None, url_name=None, **kwargs):
    """
    Mark a ViewSet method as a routable action.

    Set the `detail` boolean to determine if this action should apply to
    instance/detail requests or collection/list requests.
    """
    methods = ['get'] if (methods is None) else methods
    methods = [method.lower() for method in methods]

    assert detail is not None, (
        "@action() missing required argument: 'detail'"
    )

    def decorator(func):
        func.bind_to_methods = methods
        func.detail = detail
        func.url_path = url_path if url_path else func.__name__
        func.url_name = url_name if url_name else func.__name__.replace('_', '-')
        func.kwargs = kwargs
        return func
    return decorator


def detail_route(methods=None, **kwargs):
    """
    Used to mark a method on a ViewSet that should be routed for detail requests.
    """
    warnings.warn(
        "`detail_route` is pending deprecation and will be removed in 3.10 in favor of "
        "`action`, which accepts a `detail` bool. Use `@action(detail=True)` instead.",
        PendingDeprecationWarning, stacklevel=2
    )

    def decorator(func):
        func = action(methods, detail=True, **kwargs)(func)
        if 'url_name' not in kwargs:
            func.url_name = func.url_path.replace('_', '-')
        return func
    return decorator


def list_route(methods=None, **kwargs):
    """
    Used to mark a method on a ViewSet that should be routed for list requests.
    """
    warnings.warn(
        "`list_route` is pending deprecation and will be removed in 3.10 in favor of "
        "`action`, which accepts a `detail` bool. Use `@action(detail=False)` instead.",
        PendingDeprecationWarning, stacklevel=2
    )

    def decorator(func):
        func = action(methods, detail=False, **kwargs)(func)
        if 'url_name' not in kwargs:
            func.url_name = func.url_path.replace('_', '-')
        return func
    return decorator



from collections import OrderedDict

from django.core.paginator import Paginator
from django.db import connection
from django.utils.functional import cached_property

from jet_django.deps.rest_framework.pagination import PageNumberPagination
from jet_django.deps.rest_framework.response import Response


class CustomPaginator(Paginator):
    def count_for_postgresql(self, db_table):
        cursor = connection.cursor()
        cursor.execute('SELECT reltuples FROM pg_class WHERE relname = %s', [db_table])
        return int(cursor.fetchone()[0])

    def count_for_mysql(self, db_table):
        cursor = connection.cursor()
        cursor.execute('EXPLAIN SELECT COUNT(*) FROM `{}`'.format(db_table))
        return int(cursor.fetchone()[8])

    @cached_property
    def count(self):
        query = self.object_list.query
        result = None

        if not query.where:
            try:
                if connection.vendor == 'postgresql':
                    result = self.count_for_postgresql(query.model._meta.db_table)
                elif connection.vendor == 'mysql':
                    result = self.count_for_mysql(query.model._meta.db_table)
            except:
                pass

        if result is not None and result >= 10000:
            return result

        return super(CustomPaginator, self).count


class CustomPageNumberPagination(PageNumberPagination):
    page_size_query_param = '_per_page'
    django_paginator_class = CustomPaginator

    def get_paginated_response(self, data):
        return Response(OrderedDict([
            ('count', self.page.paginator.count),
            ('next', self.get_next_link()),
            ('previous', self.get_previous_link()),
            ('results', data),
            ('num_pages', self.page.paginator.num_pages),
            ('per_page', self.page.paginator.per_page),
        ]))



import warnings

from django.template import loader
from django.utils.deprecation import RenameMethodsBase

from . import filters, filterset
from .. import compat, utils


# TODO: remove metaclass in 2.1
class RenameAttributes(utils.RenameAttributesBase, RenameMethodsBase):
    renamed_attributes = (
        ('default_filter_set', 'filterset_base', utils.MigrationNotice),
    )
    renamed_methods = (
        ('get_filter_class', 'get_filterset_class', utils.MigrationNotice),
    )


class DjangoFilterBackend(metaclass=RenameAttributes):
    filterset_base = filterset.FilterSet
    raise_exception = True

    @property
    def template(self):
        if compat.is_crispy():
            return 'jet_django.deps.django_filters/rest_framework/crispy_form.html'
        return 'jet_django.deps.django_filters/rest_framework/form.html'

    def get_filterset(self, request, queryset, view):
        filterset_class = self.get_filterset_class(view, queryset)
        if filterset_class is None:
            return None

        kwargs = self.get_filterset_kwargs(request, queryset, view)
        return filterset_class(**kwargs)

    def get_filterset_class(self, view, queryset=None):
        """
        Return the `FilterSet` class used to filter the queryset.
        """
        filterset_class = getattr(view, 'filterset_class', None)
        filterset_fields = getattr(view, 'filterset_fields', None)

        # TODO: remove assertion in 2.1
        if filterset_class is None and hasattr(view, 'filter_class'):
            utils.deprecate(
                "`%s.filter_class` attribute should be renamed `filterset_class`."
                % view.__class__.__name__)
            filterset_class = getattr(view, 'filter_class', None)

        # TODO: remove assertion in 2.1
        if filterset_fields is None and hasattr(view, 'filter_fields'):
            utils.deprecate(
                "`%s.filter_fields` attribute should be renamed `filterset_fields`."
                % view.__class__.__name__)
            filterset_fields = getattr(view, 'filter_fields', None)

        if filterset_class:
            filterset_model = filterset_class._meta.model

            # FilterSets do not need to specify a Meta class
            if filterset_model and queryset is not None:
                assert issubclass(queryset.model, filterset_model), \
                    'FilterSet model %s does not match queryset model %s' % \
                    (filterset_model, queryset.model)

            return filterset_class

        if filterset_fields and queryset is not None:
            MetaBase = getattr(self.filterset_base, 'Meta', object)

            class AutoFilterSet(self.filterset_base):
                class Meta(MetaBase):
                    model = queryset.model
                    fields = filterset_fields

            return AutoFilterSet

        return None

    def get_filterset_kwargs(self, request, queryset, view):
        return {
            'data': request.query_params,
            'queryset': queryset,
            'request': request,
        }

    def filter_queryset(self, request, queryset, view):
        filterset = self.get_filterset(request, queryset, view)
        if filterset is None:
            return queryset

        if not filterset.is_valid() and self.raise_exception:
            raise utils.translate_validation(filterset.errors)
        return filterset.qs

    def to_html(self, request, queryset, view):
        filterset = self.get_filterset(request, queryset, view)
        if filterset is None:
            return None

        template = loader.get_template(self.template)
        context = {'filter': filterset}
        return template.render(context, request)

    def get_coreschema_field(self, field):
        if isinstance(field, filters.NumberFilter):
            field_cls = compat.coreschema.Number
        else:
            field_cls = compat.coreschema.String
        return field_cls(
            description=str(field.extra.get('help_text', ''))
        )

    def get_schema_fields(self, view):
        # This is not compatible with widgets where the query param differs from the
        # filter's attribute name. Notably, this includes `MultiWidget`, where query
        # params will be of the format `<name>_0`, `<name>_1`, etc...
        assert compat.coreapi is not None, 'coreapi must be installed to use `get_schema_fields()`'
        assert compat.coreschema is not None, 'coreschema must be installed to use `get_schema_fields()`'

        try:
            queryset = view.get_queryset()
        except Exception:
            queryset = None
            warnings.warn(
                "{} is not compatible with schema generation".format(view.__class__)
            )

        filterset_class = self.get_filterset_class(view, queryset)

        return [] if not filterset_class else [
            compat.coreapi.Field(
                name=field_name,
                required=field.extra['required'],
                location='query',
                schema=self.get_coreschema_field(field)
            ) for field_name, field in filterset_class.base_filters.items()
        ]




class MethodOverrideViewMixin(object):
    def dispatch(self, request, *args, **kwargs):
        METHOD_OVERRIDE_HEADER = 'HTTP_X_HTTP_METHOD_OVERRIDE'

        if METHOD_OVERRIDE_HEADER in request.META:
            request.method = request.META[METHOD_OVERRIDE_HEADER]

        return super().dispatch(request, *args, **kwargs)



from jet_django.admin.model_description import JetAdminModelDescription


class JetAdmin(object):
    models = []
    message_handlers = {}

    def register(self, Model, fields=None, actions=list(), hidden=False):
        self.models.append(JetAdminModelDescription(Model, fields, actions, hidden))

    def register_related_models(self):
        def model_key(x):
            return x['model']
        registered = set(map(lambda x: model_key(x.get_model()), self.models))

        for models_description in self.models:
            for item in models_description.get_related_models():
                key = model_key(item['model_info'])

                if key in registered:
                    continue

                self.register(item['model'], hidden=True)
                registered.add(key)

    def add_message_handler(self, message_name, func):
        self.message_handlers[message_name] = func

    def get_message_handler(self, message_name):
        return self.message_handlers.get(message_name)

jet = JetAdmin()



from jet_django.deps.rest_framework import views
from jet_django.deps.rest_framework.response import Response
from jet_django.admin.jet import jet
from jet_django.mixins.cors_api_view import CORSAPIViewMixin
from jet_django.permissions import HasProjectPermissions


class ModelDescriptionView(CORSAPIViewMixin, views.APIView):
    authentication_classes = ()
    permission_classes = (HasProjectPermissions,)

    def get(self, request, *args, **kwargs):
        return Response(map(lambda x: x.serialize(), jet.models))



from django.db import connection
from django.db.models.sql import Query


def get_row_number(Model, instance, where_sql, where_args, order_by_sql):
    pk = Model._meta.pk.name
    table = Model._meta.db_table

    with connection.cursor() as cursor:
        query = '''
            SELECT 
              rows.row 
            FROM 
              (
                SELECT 
                  {} as pk, 
                  ROW_NUMBER() OVER({}) AS row 
                FROM 
                  {}
                {}
              ) AS rows 
            WHERE 
              rows.pk = %s
        '''.format(
            pk,
            order_by_sql,
            table,
            where_sql
        )

        cursor.execute(query, where_args + [instance.pk])
        row = cursor.fetchone()

    if not row:
        return

    return row[0]


def get_row_siblings(Model, row_number, where_sql, where_args, order_by_sql):
    pk = Model._meta.pk.name
    table = Model._meta.db_table

    has_previous = row_number > 1
    offset = row_number - 2 if has_previous else row_number - 1
    limit = 3 if has_previous else 2

    with connection.cursor() as cursor:
        query = '''
                    SELECT {} 
                    FROM {} 
                    {}
                    {} 
                    LIMIT %s 
                    OFFSET %s
                '''.format(
            pk,
            table,
            where_sql,
            order_by_sql
        )
        cursor.execute(query, where_args + [limit, offset])
        rows = cursor.fetchall()

    if has_previous:
        next_index = 2
    else:
        next_index = 1

    if next_index >= len(rows):
        next_index = None

    if has_previous:
        prev_index = 0
    else:
        prev_index = None

    def map_row(row):
        columns = (x.name for x in cursor.description)
        return dict(zip(columns, row))

    return {
        'prev': map_row(rows[prev_index]) if prev_index is not None else None,
        'next': map_row(rows[next_index]) if next_index is not None else None
    }


def get_model_siblings(Model, instance, queryset):
    def map_ordering(x):
        asc = x[0:1] != '-'
        name = x if asc else x[1:]
        operator = 'ASC' if asc else 'DESC'
        return '{} {}'.format(name, operator)

    ordering = queryset.query.order_by
    compiler = Query(Model).get_compiler(connection=connection)
    where_query, where_args = queryset.query.where.as_sql(compiler, connection)
    where_sql = 'WHERE {}'.format(where_query) if where_query != '' else ''
    order_by = list(map(map_ordering, ordering))
    order_by_sql = 'ORDER BY {}'.format(', '.join(order_by)) if len(order_by) else ''

    row_number = get_row_number(Model, instance, where_sql, where_args, order_by_sql)

    if not row_number:
        return {}

    return get_row_siblings(Model, row_number, where_sql, where_args, order_by_sql)
